 /*-------------------------------------------------------------------*         
  *    Name: BIPLOT.SAS                                               *
  *   Title: Construct a biplot of observations and variables         *
  *          Uses IML.                                                *         
  *     Doc: http://www.math.yorku.ca/SCS/sssg/biplot.html            *
  *-------------------------------------------------------------------*         
  *  Author:  Michael Friendly            <friendly@yorku.ca>         *         
  * Created:   1 Mar 1989 13:16:36                                    *         
  * Revised:  22 Jul 1998 11:08:14                                    *         
  * Version:  1.6                                                     *         
  * 1.5  Added dimension labels, fixed problem with dim=3,            *
  *      Added colors option, Fixed problem with var=_NUM_            *
  * 1.6  Added power transformation (for log(freq))                   *
  *      Added point symbols, marker styles (interp=)                 *
  *      Made ID optional, can be char or numeric                     *
  *      Fixed bug introduced with ID                                 *
  *                                                                   *         
  *      From ``SAS System for Statistical Graphics, First Edition''  *         
  *      Copyright(c) 1991 by SAS Institute Inc., Cary, NC, USA       *         
  *-------------------------------------------------------------------*/        
                                                                                
%macro BIPLOT(                                                                  
	data=_LAST_,       /* Data set for biplot                      */         
	var =_NUM_,        /* Variables for biplot                     */         
	id  =ID,           /* Observation ID variable                  */         
	dim =2,            /* Number of biplot dimensions              */         
	factype=SYM,       /* Biplot factor type: GH, SYM, or JK       */         
	scale=1,           /* Scale factor for variable vectors        */
	power=1,           /* Power transform of response              */
	out =BIPLOT,       /* Output dataset: biplot coordinates       */         
	anno=BIANNO,       /* Output dataset: annotate labels          */         
	xanno=dim1,
	yanno=dim2,
	zanno=dim3,
	std=MEAN,          /* How to standardize columns: NONE|MEAN|STD*/
	colors=BLUE RED,   /* Colors for OBS and VARS                  */
	symbols=none none, /* Symbols for OBS and VARS                 */
	interp=none vec,   /* Markers/interpolation for OBS and VARS   */          
	pplot=NO,          /* Produce printer plot?                    */
	gplot=YES,
	haxis=,            /* AXIS statement for horizontal axis       */
	vaxis=,            /* and for vertical axis- use to equate axes */
	name=biplot);        
                                                                                
%let std=%upcase(&std);                                                 
%let factype=%upcase(&factype);                                                 
      %if &factype=GH  %then %let p=0;                                          
%else %if &factype=SYM %then %let p=.5;                                         
%else %if &factype=JK  %then %let p=1;                                          
%else %do;                                                                      
   %put BIPLOT: FACTYPE must be GH, SYM, or JK. "&factype" is not valid.;       
   %goto done;                                                                  
   %end;                                                                        
%if %upcase("&var") ^= "_NUM_" %then %let var={&var};
%if &data=_LAST_ %then %let data=&syslast;
                                                                                
proc iml;                                                                       
start biplot(y,id,vars,out, g, scale);                                      
   N = nrow(Y);                                                                 
   P = ncol(Y);                                                                 
   %if &std = NONE                                                              
       %then Y = Y - Y[:] %str(;);             /* remove grand mean */          
       %else Y = Y - J(N,1,1)*Y[:,] %str(;);   /* remove column means */        
   %if &std = STD %then %do;                                                    
      S = sqrt(Y[##,] / (N-1));                                                 
      Y = Y * diag (1 / S );                                                    
   %end;                                                                        
                                                                                
   *-- Singular value decomposition:                                            
        Y is expressed as U diag(Q) V prime                                     
        Q contains singular values, in descending order;                        
   call svd(u,q,v,y);                                                           
                                                                                
   reset fw=8 noname;                                                           
   percent = 100*q##2 / q[##];                                                  
   cum = cusum(percent);                                                           
   c1={'Singular Values'};                                                      
   c2={'Percent'};                                                              
   c3={'Cum % '};                                                               
   Print "Singular values and variance accounted for",,                         
         q       [colname=c1 format=9.4 ]                                       
         percent [colname=c2 format=8.2 ]                                       
         cum     [colname=c3 format=8.2 ];                                      
                                                                                
   d = &dim ;
	*-- Assign macro variables for dimension labels;
	lab = '%let p' + char(t(1:d),1) + '=' + left(char(percent[t(1:d)],8,1)) + ';';
	call execute(lab);
	/*                                                                  
	call execute('%let p1=', char(percent[1],8,1), ';');
	call execute('%let p2=', char(percent[2],8,1), ';');
	if d > 2 then
	call execute('%let p3=', char(percent[3],8,1), ';');
	*/
   *-- Extract first  d  columns of U & V, and first  d  elements of Q;         
   U = U[,1:d];                                                                 
   V = V[,1:d];                                                                 
   Q = Q[1:d];                                                                  
                                                                                
   *-- Scale the vectors by QL, QR;                                             
   * Scale factor 'scale' allows expanding or contracting the variable          
     vectors to plot in the same space as the observations;                     
   QL= diag(Q ## g );                                                       
   QR= diag(Q ## (1-g));                                                    
   A = U * QL;                                                                  
   B = V * QR;
	ratio = max(sqrt(A[,##])) /  max(sqrt(B[,##]));
	print 'OBS / VARS ratio:' ratio 'Scale:' scale;
	if scale=0 then scale=ratio;                                                         
   B = B # scale;                                                          
   OUT=A // B;                                                                  
                                                                                
   *-- Create observation labels;                                               
   id = id // vars`;                                                            
   type = repeat({"OBS "},n,1) // repeat({"VAR "},p,1);                         
   id  = concat(type, id);                                                      
                                                                                
   factype = {"GH" "Symmetric" "JK"}[1 + 2#g];                              
   print "Biplot Factor Type", factype;                                         
                                                                                
   cvar = concat(shape({"DIM"},1,d), char(1:d,1.));                             
   print "Biplot coordinates",                                                  
         out[rowname=id colname=cvar f=9.4];                                          
   %if &pplot = YES %then %do;                                                       
   call pgraf(out[,{1 2}],substr(id,5),'Dimension 1', 'Dimension 2', 'Biplot');         
      %end;                                                                         
   create &out  from out[rowname=id colname=cvar];                              
   append from out[rowname=id];                                                 
finish;                                                                         

start power(x, pow);
		if pow=1 then return(x);
		if any(x <= 0) then x = x + ceil(min(x)+.5);
		if abs(pow)<.001 then xt =  log(x);
			else xt = ((x##pow)-1) / pow;
	return (xt);
	finish;

	/*--- Main routine */
                                                                                
   use &data;                                                                   
   read all var &var into y[ c=vars ];
   %if &id = %str() %then %do;
      id=compress(char(1:nrow(xy),4))`;
   %end;
   %else %do;
      read all var{&id} into id;
   %end;
*   read all var &var into y[colname=vars rowname=&id];                         

	%if &power ^= 1 %then %do;
		y = power(y, &power);
		%end;

   scale = &scale;                                                              
   run biplot(y, id,vars,out, &p, scale );                                  
   quit;                                                                        
                                                                                
 /*----------------------------------*                                          
  |  Split ID into _TYPE_ and _NAME_ |                                          
  *----------------------------------*/                                         

data &out;                                                                      
   set &out;                                                                    
   drop id;                                                                     
   length _type_ $3 _name_ $16;                                                 
   _type_ = substr(id,1,3);                                                     
   _name_ = substr(id,5);                                                       
   label                                                                        
   %do i=1 %to &dim;                                                            
       dim&i = "Dimension &i (&&p&i%str(%%))"                                                   
       %end;                                                                    
   ;                                                                            
 /*--------------------------------------------------*                          
  | Annotate observation labels and variable vectors |                          
  *--------------------------------------------------*/                         
	%*-- Assign colors and symbols;
	%let c1= %scan(&colors,1);                                                      
	%let c2= %scan(&colors,2);
	%if &c2=%str() %then %let c2=&c1;                                                   

	%let v1= %upcase(%scan(&symbols,1));                                                      
	%let v2= %upcase(%scan(&symbols,2));
	%if &v2=%str() %then %let v2=&v1;                                                   

	%let i1= %upcase(%scan(&interp,1));                                                      
	%let i2= %upcase(%scan(&interp,2));
	%if &i2=%str() %then %let i2=&i1;                                                   

data &anno;                                                                     
   set &out;                                                                    
   length function color $8 text $16;                                                 
   xsys='2'; ysys='2'; %if &dim > 2 %then %str(zsys='2';);                                               
   text = _name_;                                                               
                                                                                
   if _type_ = 'OBS' then do;         /* Label observations (row points) */             
      color="&c1";                                                            
		if "&i1" = 'VEC' then link vec;                                                            
      x = &xanno; y = &yanno;                                                       
      %if &dim > 2 %then %str(z = &zanno;);
		%if &v1=NONE %then                                     
      	%str(position='5';);
		%else %do;
      if dim1 >=0                                                               
         then position='>';             /* rt justify         */                
         else position='<';             /* lt justify         */                
			%end;                                                             
      function='LABEL   '; output;                                              
      end;                                                                      
                                                                                
   if _type_ = 'VAR' then do;           /* Label variables (col points) */                
      color="&c2";
		if "&i2" = 'VEC' then link vec;                                                            
      x = &xanno; y = &yanno;                
      if dim1 >=0                                                               
         then position='6';             /* down justify       */                
         else position='2';             /* up justify         */                
      function='LABEL   '; output;      /* variable name      */                
      end;                                                                      
		return;

vec:                    /* Draw line from the origin to point */
      x = 0; y = 0;                
      %if &dim > 2 %then %str(z = 0;);                                          
      function='MOVE'    ; output;                                              
      x = &xanno; y = &yanno;                
      %if &dim > 2 %then %str(z = &zanno;);                                       
      function='DRAW'    ; output;                                              
		return;

%if &gplot = YES %then %do;
	%if &i1=VEC %then %let i1=NONE;
	%if &i2=VEC %then %let i2=NONE;
	%let legend=nolegend;

	%let warn=0;
   %if %length(&haxis)=0 %then %do;
		%let warn=1;
		axis2 offset=(1,5) ;
      %let haxis=axis2;
      %end;
   %if %length(&vaxis)=0 %then %do;
		%let warn=1;
		axis1 offset=(1,5) label=(a=90 r=0);
      %let vaxis=axis1;
      %end;

	proc gplot data=&out &GOUT;
		plot dim2 * dim1 = _type_/ 
			anno=&anno frame &legend
			href=0 vref=0 lvref=3 lhref=3
			vaxis=&vaxis haxis=&haxis
			vminor=1 hminor=1
			name="&name" des="Biplot of &data";
		symbol1 v=&v1 c=&c1 i=&i1;
		symbol2 v=&v2 c=&c2 i=&i2;
	run; quit;
	%if &warn %then %do;
		%put WARNING:  No VAXIS= or HAXIS= parameter was specified, so the biplot axes have not;
		%put WARNING:  been equated.  This may lead to incorrect interpretation of distance and;
		%put WARNING:  angles.  See the documentation.;
		%end;
	goptions reset=symbol;
	%end;  /* %if &gplot=YES */

%done:                                                                          
%mend BIPLOT;                                                                   
