 /*-------------------------------------------------------------------*
  *    Name: CONTOUR SAS                                              *
  *   Title: IML macro to plot elliptical contours for X, Y data      *
  *     Ref: IML User's Guide, version 5 Edition                      *
  *-------------------------------------------------------------------*
  *  Author:  Michael Friendly            <friendly@VM1.YorkU.ca>     *
  * Created:  8 Jun 1988 12:33:21                                     *
  * Revised:  25 Sep 1991 13:08:30                                    *
  * Version:  2.0                                                     *
  *      From ``SAS System for Statistical Graphics, First Edition''  *
  *      Copyright(c) 1991 by SAS Institute Inc., Cary, NC, USA       *
  *                                                                   *
  *-------------------------------------------------------------------*/
 
%macro CONTOUR(
       data=_LAST_,           /* input data set                   */
       x=,                    /* X variable                       */
       y=,                    /* Y variable                       */
       group=,                /* Group variable (optional)        */
       pvalue= .5,            /* Confidence coefficient (1-alpha) */
       std=STDERR,            /* error bar metric: STD or STDERR  */
       points=40,             /* points on each contour           */
       all=NO,                /* include contour for total sample?*/
       out=CONTOUR,           /* output data set                  */
       plot=YES,              /* plot the results?                */
       i=none,                /* SYMBOL statement interpolate opt */
       name=CONTOUR,          /* Name for graphic catalog entry   */
       colors=RED GREEN BLUE BLACK PURPLE YELLOW BROWN ORANGE,
       symbols=+ square star -     plus   :      $     = );
 
%let all = %upcase(&all);
%if &x=%str() or &y=%str() %then %do;
   %put CONTOUR: X= and Y= variables must be specified;
   %goto DONE;
   %end;
 
proc iml;
start ellipse(c, x, y, npoints, pvalues, formean);
  /*----------------------------------------------------------------*
   |  Computes elliptical contours for a scatterplot                |
   |   C       returns the contours as consecutive pairs of columns |
   |   X,Y     coordinates of the points                            |
   |   NPOINTS scalar giving number of points around a contour      |
   |   PVALUES column vector of confidence coefficients             |
   |   FORMEAN 0=contours for observations, 1=contours for means    |
   *----------------------------------------------------------------*/
 
   xx = x||y;
   n  = nrow(x);
   *-- Correct for the mean --;
   mean = xx[+,]/n;
   xx = xx - mean @ j(n,1,1);
 
   *-- Find principal axes of ellipses --;
   xx = xx` * xx / (n-1);
   print 'Variance-Covariance Matrix',xx;
   call eigen(v, e, xx);
 
   *-- Set contour levels --;
   c =  2*finv(pvalues,2,n-1,0);
   if formean=1 then c = c / (n-1) ;
   print 'Contour values',pvalues c;
   a = sqrt(c*v[ 1 ] );
   b = sqrt(c*v[ 2 ] );
 
   *-- Parameterize the ellipse by angles around unit circle --;
   t = ( (1:npoints) - {1}) # atan(1)#8/(npoints-1);
   s = sin(t);
   t = cos(t);
   s = s` * a;
   t = t` * b;
 
   *-- Form contour points --;
   s = ( ( e*(shape(s,1)//shape(t,1) )) +
         mean` @ j(1,npoints*ncol(c),1) )` ;
   c = shape( s, npoints);
   *-- C returned as NCOL pairs of columns for contours--;
finish;
 
start dogroups(x, y, gp, pvalue);
   d  = design(gp);
   %if &all=YES %then %do;
      d = d || j(nrow(x),1,1);
   %end;
   do group = 1 to ncol(d);
      Print group;
      *-- select observations in each group;
      col = d[, group ];
      xg  = x[ loc(col), ];
      yg  = y[ loc(col), ];
      *-- Find ellipse boundary ;
      run ellipse(xyg,xg,yg,&points, pvalue, 0 );
      nr = nrow(xyg);
 
      *-- Output contour data for this group;
      cnames = { X Y PVALUE GP };
      do c=1 to ncol(pvalue);
         col=(2*c)-1 : 2*c ;
         xygp = xyg[,col] || j(nr,1,pvalue[c]) || j(nr,1,group);
         if group=1 & c=1
            then create contour from xygp [colname=cnames];
         append from xygp;
      end;
   end;
finish;
 
*-- Get input data: X, Y, GP;
   use &data;
   read all var {&x}     into x [colname=lx];
   read all var {&y}     into y [colname=ly];
%if &group ^= %str() %then %do;
   read all var {&group} into gp [colname=lg] ;
   %end;
%else %do;
   gp = j(nrow(x),1,1);
   %end;
   close &data;
 
*-- Find contours for each group;
   run dogroups(x, y, gp, { &pvalue} );
quit;
 
 /*-----------------------------------*
  |  Plot the contours using ANNOTATE |
  *-----------------------------------*/
data contour;
   set contour ;
   by gp pvalue notsorted;
   length function color $8;
   xsys='2'; ysys='2';
   if first.pvalue then function='POLY';
                   else function='POLYCONT';
   color=scan("&colors",gp);
   line = 5;
run;
 /*----------------------------*
  |  Crosses at Mean +- StdErr |
  *----------------------------*/
proc summary data=&data nway;
   class &group;
   var &x &y;
   output out=sumry mean=mx my &std=sx sy;
run;
data bars;
   set sumry end=eof;
   %if &group ^= %str() %then %str(by &group;);
   length function color $8;
   retain g 1;
   drop _freq_ _type_ mx my sx sy g;
   xsys='2'; ysys='2';
 
   color=scan("&colors",g);
   line=3;
   x = mx-sx; y=my;      function='MOVE'; output;
   x = mx+sx;            function='DRAW'; output;
   x = mx   ; y=my-sy;   function='MOVE'; output;
              y=my+sy;   function='DRAW'; output;
 
 *-- Write group label (convert numeric &group to character);
   %if &group ^= %str() %then %do;
      length text $16;
      text = left(&group);
      position='3';
      size = 1.4;
      x = mx+.2*sx   ; y=my+.2*sy;
      function='LABEL'; output;
   %end;
   if eof then call symput('NGROUP',put(g,best.));
   %if &group ^= %str() %then %do;
      if last.&group then g+1;
   %end;
run;
 
data &out;
   set contour bars;
 
%if &group = %str() %then %let group=1;
%if %upcase(&plot)=YES %then %do;
   %gensym(n=&ngroup, h=1.2, i=&i, colors=&colors, symbols=&symbols );
   proc gplot data=&data &GOUT ;
      plot &y * &x = &group /
           annotate=&out
           nolegend frame
           vaxis=axis1 vminor=0
           haxis=axis2 hminor=0
           name="&name";
      axis1 offset=(3) value=(h=1.5) label=(h=1.5 a=90 r=0);
      axis2 offset=(3) value=(h=1.5) label=(h=1.5);
      run;
%end;
%done: ;
%mend contour;
 
 /*----------------------------------------------------*
  |  Macro to generate SYMBOL statement for each GROUP |
  *----------------------------------------------------*/
%macro gensym(n=1, h=1.5, i=none,
              symbols=%str(- + : $ = X _ Y),
              colors=BLACK RED GREEN BLUE BROWN YELLOW ORANGE PURPLE);
    %*-- note: only 8 symbols & colors are defined;
    %*--    revise if more than 8 groups (recycle);
  %local chr col k;
  %do k=1 %to &n ;
     %let chr =%scan(&symbols, &k,' ');
     %let col =%scan(&colors, &k, ' ');
     symbol&k h=&h v=&chr c=&col i=&i;
  %end;
%mend gensym;
