 /*-------------------------------------------------------------------*
  *    Name: CORRESP.SAS                                              *
  *   Title: Correspondence analysis of contingency tables            *
  *                                                                   *
  *-------------------------------------------------------------------*
  *  Author:  Michael Friendly            <friendly@yorku.ca>         *
  * Created:  19 Jan 1990 15:23:09                                    *
  * Revised:   3 May 1998 12:10:08                                    *
  * Version:  1.2                                                     *
  * 1.2  Added dim parameter and colors                               *
  *      From ``SAS System for Statistical Graphics, First Edition''  *
  *      Copyright(c) 1991 by SAS Institute Inc., Cary, NC, USA       *
  *                                                                   *
  *-------------------------------------------------------------------*/
%macro CORRESP(
     data=_LAST_,        /* Name of input data set               */
     var=,               /* Column variables                     */
     id=,                /* ID variable: row labels              */
     dim =2,             /* Number of CA dimensions              */         
     out=COORD,          /* output data set for coordinates      */
     anno=LABEL,         /* name of annotate data set for labels */
     rowht=,             /* height for row labels                */
     colht=,             /* height for col labels                */
     colors=BLUE RED,    /* Colors for rows and cols                  */
     pplot=NO,         /* Produce printer plot?                    */
	gplot=YES,
	pos=5 5,
	symbols=none,
	interp=none,
	haxis=,            /* AXIS statement for horizontal axis       */
	vaxis=,            /* and for vertical axis- use to equate axes */
	name=corresp        
     ); 
 /*------------------------------------------*
  |  IML routine for Correspondence Analysis |
  *------------------------------------------*/
%let pplot=%upcase(&pplot);
%let gplot=%upcase(&gplot);

proc iml;
start corresp(f,rowid,vars);
   I = nrow(F);
   J = ncol(F);
   R = F[,+];                   * Row totals;
   C = F[+,];                   * Col totals;
   N = F[+];                    * Grand total;
   E  = R * C / N;              * Expected frequencies;
   D = (F - E) / sqrt(E);       * Standardized deviates;
   D = D / sqrt( N );
   DPD = D` * D;
   Inertia = trace(DPD);
   Chisq = N * Inertia;         * Total chi-square;
   DF = (I-1)*(J-1);
   reset noname;
   C1={'ChiSq'};
   C2={'DF'};
   Print 'Overall Association', CHISQ[colname=C1]
                                DF[colname=C2 format=6.0];
 
   call eigen(values, vectors, dpd);
   k = min(I,J)-1;                * number of non-zero eigenvalues;
   values  = values[1:k];
   cancorr = sqrt(values);        * singular values = Can R;
   chisq = n * values ;           * contribution to chi-square;
   percent = 100* values / inertia;
      *-- Cumulate by multiplying by lower triangular matrix of 1s;
   tri= (1:k)`*repeat(1,1,k)  >= repeat(1,k,1)*(1:k) ;
   cum = tri*percent;
   C1={'Singular Values'};
   C2={'Inertias'};
   C3={'Chi-Squares'};
   C4={'Percent'};
   C5={' Cum % '};
   print 'Singular values, Inertia, and Chi-Square Decomposition',,
         cancorr [colname=C1 format=9.4]
         values  [colname=C2 format=9.4]
         chisq   [colname=C3 format=9.3]
         percent [colname=C4 format=8.2]
         cum     [colname=C5 format=8.2];
 
   d = &dim;
   L = values[1:d];
   U = vectors[,1:d];
   Y = diag(1/sqrt(C/N)) * U * diag(sqrt(L));
   X = diag(N/R) * (F / N) * Y * diag(sqrt(1/L));

	c1 = 'DIM1' : ('DIM'+char(d,1.));
   Print 'Row Coordinates'   , X [Rowname=RowId Colname=C1];
   Print 'Column Coordinates', Y [Rowname=Vars  Colname=C1];

	*-- Assign macro variables for dimension labels;
	lab = '%let p' + char(t(1:d),1) + '=' + left(char(percent[t(1:d)],8,1)) + ';';
	print lab;
	call execute(lab);
 
   OUT = X // Y;
   ID  = RowId // Vars`;
   %if &pplot = YES %then %do;                                                       
   Call PGRAF(out[,{1 2}],ID,'Dimension 1', 'Dimension 2', 'Row/Col Association');
      %end;                                                                         

   TYPE = repeat({"OBS "},I,1) // repeat({"VAR "},J,1);
   ID  = concat(TYPE, ID);
   Create &out  from OUT[rowname=ID colname={"Dim1" "Dim2"}];
   Append from OUT[rowname=ID];
Finish;
 
use &data;
read all var {&var} into f [rowname=&id colname=vars];
 
run corresp(f,&id,vars);
quit;

 /*----------------------------------*
  |  Split ID into _TYPE_ and _NAME_ |
  *----------------------------------*/
data &out;
   set &out;
   drop id;
   length _type_ $3 _name_ $16;
   _type_ = substr(id,1,3);
   _name_ = substr(id,5);
   label                                                                        
   %do i=1 %to &dim;                                                            
       dim&i = "Dimension &i (&&p&i%str(%%))"                                                   
       %end;                                                                    
   ;                                                                            
proc print data=&out;
   id _type_ _name_;

 /*--------------------------------------------------*
  | Annotate row and column labels                   |
  *--------------------------------------------------*/
	%*-- Assign colors / positions;
	%let c1= %scan(&colors,1);                                                      
	%let c2= %scan(&colors,2);
	%if &c2=%str() %then %let c2=&c1;
	%let p1 = %scan(&pos,1,%str( ));
	%let p2 = %scan(&pos,2,%str( ));
	%if &p2=%str() %then %let p2=&p1;
                                                
%label(data=&out, x=dim1, y=dim2, text=_name_, size=&rowht, color="&c1",
	subset=_type_='OBS', pos=&p1, out=lab1, len=16);
%label(data=&out, x=dim1, y=dim2, text=_name_, size=&colht, color="&c2",
	subset=_type_='VAR', pos=&p2, out=lab2);
data &anno;
	set lab1 lab2;

%if &gplot = YES %then %do;
%if %length(&vaxis)=0 %then %do;
	%let vaxis=axis98;
	%put WARNING:  You should use an AXISn statement and specify VAXIS=AXISn to equate axis units and length;
   axis98 label=(a=90);
	%end;
%if %length(&haxis)=0 %then %do;
	%let haxis=axis99;
	%put WARNING:  You should use an AXISm statement and specify HAXIS=AXISm to equate axis units and length;
   axis99 offset=(2);
	%end;

	%*-- Assign symbols and interpolations;
	%let s1= %scan(&symbols,1);                                                      
	%let s2= %scan(&symbols,2);
	%if &s2=%str() %then %let s2=&s1;
	%let i1= %scan(&interp,1);                                                      
	%let i2= %scan(&interp,2);
	%if &i2=%str() %then %let i2=&i1;

proc gplot data=&out  ;
   plot dim2 * dim1 = _type_
        / anno=&anno frame nolegend
          vaxis=&vaxis haxis=&haxis vminor=1 hminor=1;
   symbol1 v=&s1 i=&i1 l=33 c=&c1;
   symbol2 v=&s2 i=&i2 l=20 c=&c2;
run;quit;

%end;	/* %if &gplot = YES */

%macro ignore;
data &anno;
   set &out;
   length function text $16;
	drop _name_;
   xsys='2'; ysys='2';
   text = _name_;
*   style='DUPLEX';
   x = dim1; y = dim2;
 
   if _type_ = 'OBS' then do;
		%if %length(&rowht) %then %do;
	      size= &rowht ;
			%end;
      color="&c1";                                                            
      position='5';
      function='LABEL   '; output;
      end;
 
   if _type_ = 'VAR' then do;
		%if %length(&colht) %then %do;
	      size= &colht ;
			%end;
      color="&c2";                                                            
      if dim1 >=0
         then position='6';             /* left justify       */
         else position='4';             /* right justify      */
      function='LABEL   '; output;
      end;
%mend;
%mend CORRESP;
