 /*-------------------------------------------------------------------*
  * LOWESS SAS     Locally weighted robust scatterplot smoothing      *
  *                                                                   *
  *-------------------------------------------------------------------*
  *  Author:  Michael Friendly            <friendly@yorku.ca>         *
  * Created:  21 Apr 1989 09:21:55                                    *
  * Revised:  Tue May 28 10:14:51 EDT 1996                            *
  * Version:  1.6                                                     *
  *     Add:  Filter out missing data                                 *
  *           Fit local quadratic as well as linear smooth            *
  *           Output annotate= data set                               *
  *      From ``SAS System for Statistical Graphics, First Edition''  *
  *      Copyright(c) 1991 by SAS Institute Inc., Cary, NC, USA       *
  *                                                                   *
  *-------------------------------------------------------------------*/
%macro LOWESS(
       data=_LAST_,        /* name of input data set            */
       out=SMOOTH,         /* name of output data set           */
       x = X,              /* name of independent variable      */
       y = Y,              /* name of Y variable to be smoothed */
       id=,                /* optional row ID variable          */
       f = .50,            /* lowess window width               */
       p = 1,              /* 1=linear fit, 2=quadratic         */
       iter=2,             /* total number of iterations        */
       plot=NO,            /* draw the plot?                    */
       gplot=NO,           /* draw the plot?                    */
       pplot=NO,           /* draw a printer plot?              */
       symbol=circle,      /* plotting symbol for points        */
       htext=1.5,
       hsym=1.5,
       colors=BLACK RED,   /* colors for points and smooth      */
       outanno=,
       silent=NO,
       name=LOWESS);       /* name for graphic catalog entry    */

proc sort data=&data out=_sorted_;
   by &x;
   where ((&x ^= .) & (&y ^= .));
 
proc iml;
start WLS( X, Y, W, B, I );      *-- Weighted least squares;
   x = j(nrow(x), 1, 1) || x;
   xpx = x` * diag( w ) * x;
   xpy = x` * diag( w ) * y;
   if abs(det(xpx)) > .00001
      then b   = inv(xpx) * xpy;
      else do;
         b = (y[loc(w^=0)])[:] // j(ncol(x)-1,1,0);
         print 'Singular matrix for observation', I;
      end;
finish;
 
start MEDIAN( W, M);         * calculate median ;
   n = nrow( W );
   R = rank( W );
   i = int((n+1)/2);
   i =  i || n-i+1;
   M = W[ R[i] ];
   M = .5 # M[+];
   finish;
 
start robust( r, wts);       * calculate robustness weights;
    run median(abs(R), M);
    W = R / (6 # M);         * bisquare function;
    wts = (abs(W) < 1) # (1 - W##2) ## 2;
   finish;
 
start lowess( x, y, f, steps, yhat, res, delta);
   n = nrow(X);
   if n < 2 then do;
      yhat = y;
      return;
      end;
   q = round( f * n);      * # nearest neighbors;
   q = min( q, n );        * if f>1;
 
   res  = y;
   yhat = J(n,1,0);
   delta= J(n,1,1);        * robustness weights;
   if steps <= 0 then steps=1;
   do it = 1 to steps;
      do i = 1 to n;
         dist = abs(x[,1] - x[i,1]);  * distance to each other pt;
         r = rank( dist );
         s = r; s[r]=1:n;
         near =  s[1:q] ;             * find the q nearest;
         nx = x [ near,];
         ny = y [ near ];
         d  = dist[ near[q] ];        * distance to q-th nearest;
         if f > 1 then d = d#f;       * handle f>1;
         if d > 0 then do;
            u = abs( nx[,1] - x[i,1] ) / d ;
            wts = (u < 1) # (1 - u##3) ## 3; * neighborhood wts;
            wts = delta[ near ] # wts;
            if sum(wts[2:q]) > .0001 then do;
               run wls( nx, ny, wts, b, i );
               yhat[i] = (1 || x[i,]) * b;      * smoothed value;
               end;
            else yhat[i] = y[i];
         end;
         else do;
            yhat[i] = ny [+] /q;
         end;
      end;
      res = y - yhat;
      run robust(res,delta);
   end;
   finish;
 
*-- Main routine;
  use _sorted_;
   %if &id = %str() %then %do;
      read all var{&x &y} into xy[ c=vars ];
      id='OBS'+char(1:nrow(xy),3)`;
   %end;
   %else %do;
      read all var{&x &y} into xy[ c=vars r=&id ];
      id = &id;
   %end;
  close _sorted_;
 
  x = xy[,1];
  y = xy[,2];
  if &p = 2 then x = x || x##2;
  run lowess(x, y, &f, &iter, yhat, res, weight);
 
  xyres =x[,1] || y || yhat || res || weight;
  cname = vars || {"_YHAT_" "_RESID_" "_WEIGHT_" };
  %if &silent = NO %then %do;
		print "Lowess parameters: f=&f p=&p";
		print "Data, smoothed fit, residuals and weights",
				xyres[ colname=cname r=id ];
  %end;
 
*-- Output results to data set &out ;
  xys =    yhat || res || weight;
  cname = {"_YHAT_" "_RESID_" "_WEIGHT_" };
 
  create &out from xys [ colname=cname ];
     append from xys;
quit;
 /*--------------------------------------------*
  | Merge data with smoothed results.          |
  | (In a data step to retain variable labels) |
  *--------------------------------------------*/
data &out;
   merge _sorted_(keep=&x &y &id)
         &out ;
   label _yhat_ = "Smoothed &y"
         _weight_='Lowess weight';
 
%let c1 = %scan(&colors &colors, 1);
%let c2 = %scan(&colors &colors, 2);

%if %upcase(&PLOT)=YES or %upcase(&GPLOT)=YES %then %do;
proc gplot data=&out &GOUT ;
  plot &y     * &x = 1
       _yhat_ * &x = 2
       / overlay frame
         vaxis=axis1 haxis=axis2
         name="&name" ;
  symbol1 v=&symbol  h=&hsym  i=none c=&c1;
  symbol2 v=none i=join c=&c2;
  axis1   label=(h=&htext a=90 r=0) value=(h=&htext);
  axis2   label=(h=&htext)          value=(h=&htext);
  run;
%end;
 
%if %upcase(&PPLOT)=YES %then %do;
proc plot data=&out ;
  plot &y     * &x = '*'
       _yhat_ * &x = '@' / overlay ;
run;
%end;

%*-- Create an output annotate dataset? ---;
%if %length(&outanno) > 0 %then %do;
data &outanno;
   set &out(keep=&id &x _yhat_);
	drop _yhat_;
	retain xsys ysys '2';
	length function color $8;
	if _n_=1 then function='MOVE    ';
		else function='DRAW    ';
	x = &x;
	y = _yhat_;
	color = upcase("&c2");
%end;

%mend LOWESS;
