 /*-------------------------------------------------------------------*
  *    Name: outlier.sas                                              *
  *   Title: Robust multivariate outlier detection                    *
  *     Doc: http://www.math.yorku.ca/SCS/sssg/outlier.html           *
  *                                                                   *
  * Macro to calculate robust Mahalanobis distances for each          *
  * observation in a dataset. The results are robust in that          *
  * potential outliers do not contribute to the distance of any       *
  * other observations.                                               *
  *                                                                   *
  * The macro makes one or more passes through the data. Each         *
  * pass assigns 0 weight to observations whose DSQ value             *
  * has prob < PVALUE. The number of passes should be determined      *
  * empirically so that no new observations are trimmed on the        *
  * last pass.                                                        *
  *-------------------------------------------------------------------*
  *  Author:  Michael Friendly            <friendly@YorkU.ca>         *
  * Created:  16 Jan 1989 18:38:18                                    *
  * Revised:  13 Apr 2000 09:14:37                                    *
  * Version:  1.4                                                     *
  * 1.2- Added GPLOT option to produce graphic output (uses %LABEL)   *
  *      and PPLOT option to suppress proc plot version               *
  * 1.3- Added check for missing data (remove them)                   *
  *    - Added support for abbreviated VAR= lists                     *
  *    - Plot ID labels in printer plot (if &sysver>6.07)             *
  *    - Changed default pvalue to 0.05                               *
  *    - Label outliers with _CASE_ number if no ID= supplied         *
  * 1.4- Fixed bug when LAST observation contains missing values      *
  *      (thanks to Wim Aerts <wim.aerts@sbx.sas.com>)                *
  *      From ``SAS System for Statistical Graphics, First Edition''  *
  *      Copyright(c) 1991 by SAS Institute Inc., Cary, NC, USA       *
  *                                                                   *
  *-------------------------------------------------------------------*/
%macro outlier(
         data=_LAST_,      /* Data set to analyze            */
         var=_NUMERIC_,    /* input variables                */
         id=,              /* ID variable for observations   */
         out=CHIPLOT,      /* Output dataset for plotting    */
         pvalue=.05,       /* Prob < pvalue --> weight=0     */
         passes=2,         /* Number of passes               */
         print=YES,        /* Print OUT= data set?           */
			pplot=YES,
			gplot=YES,
			symbol=dot,
			name=outlier,
			gout=gseg);
 
*-- Parse variables list if it contains special lists;
%let var=%upcase(&var);
%if %index(&var,-) > 0 or "&var"="_NUMERIC_" %then %do;
 data _null_;
 set &data (obs=1);
        %*  convert shorthand variable list to long form;
     length _vname_ $ 8 _vlist_ $ 200;
     array _xx_ &var;
     _vname_ = ' ';
     do over _xx_;
        call vname(_xx_,_vname_);
        _vlist_ = trim(_vlist_)|| ' ' || trim(_vname_);
     end;
	  put "NOTE: The VAR=&VAR list translates to: VAR=" _vlist_;
     call symput( 'VAR', trim(_vlist_) );
 RUN;
%end;

 /*-------------------------------------------------------*
  | Add WEIGHT variable. Determine number of observations |
  | and variables, and create macro variables.            |
  | Delete missing observations.                          |
  *-------------------------------------------------------*/
data in;
   set &data end=lastobs;
	retain _nobs_ _nmiss_ 0;
	drop _nmiss_ _nobs_;
   array invar{*} &var;
   _weight_ = 1;               /* Add weight variable */
	_miss_ = nmiss(of &var);
	if _miss_>0 then
		_nmiss_+1;
	else _nobs_+1;
   if ( lastobs ) then do;
      call symput('NOBS', trim(left(put(_nobs_,8.))) );
      call symput('NVAR', left(put(dim(invar),3.)) );
		if _nmiss_>0 then 
		put 'WARNING: ' _nmiss_ 'of ' _n_ ' observations were omitted due to missing values.';
      end;
	if _miss_>0 then delete;

run;

options nonotes;
%do pass = 1 %to &PASSES;
   %if &pass=1 %then %let in=in;
               %else %let in=trimmed;
 /*--------------------------------------------------------------*
  | Transform variables to scores on principal components.       |
  | Observations with _WEIGHT_=0 are not used in the calculation,|
  | but get component scores based on the remaining observations.|
  *--------------------------------------------------------------*/
   proc princomp std noprint data=&in out=prin;
      var &var;
      freq _weight_;
 
 /*-----------------------------------------------------------*
  | Calculate Mahalanobis D**2 and its probability value. For |
  | standardized principal components, D**2 is just the sum   |
  | of squares. Output potential outliers to separate dataset.|
  *-----------------------------------------------------------*/
   data out1    (keep=_pass_ _case_ &id dsq prob)
        trimmed (drop=_pass_ _case_ );
      set prin ;
      _pass_ = &pass;
      _case_ = _n_;
 
      dsq = uss(of prin1-prin&nvar);    /* Mahalanobis D**2 */
      prob = 1 - probchi(dsq, &nvar);
      _weight_ = (prob > &pvalue);
      output trimmed;
      if _weight_ = 0 then do;
         output out1   ;
         end;
   run;
   proc append base=outlier data=out1;
	run;
%end; /* %do passes */

   proc print data=outlier;
		id _pass_;
		by _pass_;
   title2 'Observations trimmed in calculating Mahalanobis distance';
 /*------------------------------------------*
  | Prepare for Chi-Square probability plot. |
  *------------------------------------------*/
proc sort data=trimmed;
   by dsq;
	run;
options notes;
data &out;
   set trimmed;
   drop prin1 - prin&nvar;
   _weight_ = prob > &pvalue;
   expected = 2 * gaminv(_n_/(&nobs+1), (&nvar/2));
   label dsq     ='Squared Distance'
         expected='Chi-square quantile';
 
%if %upcase(&print)=YES %then %do;
proc print data=&out;
   %if &id ^=%str() %then
   %str(id &id;);
	var _weight_ dsq prob &var;
   title2 'Possible multivariate outliers have _WEIGHT_=0';
%end;
 
%if %upcase(&PPLOT) = YES 
	%then %do;
	%if &ID = %str() 
		%then %let symbol='*';
		%else %do;
			/* Use proc plot point labels, but just for outliers */
			%if &sysver > 6.07
				%then %do;
				data &out;
					set &out;
					if _weight_=0 then _id_ = &id;			
				%let symbol = '*' $ _id_;
				%end; 
			/* Not available before 6.07 */
				%else %let symbol=&id;
			%end;
%end;
		
proc plot data=&out;
   plot dsq      * expected = &symbol
        expected * expected = '.'   /overlay hzero vzero;
title2 'Chi-Squared probability plot for multivariate outliers';
run;

%if %upcase(&GPLOT) = YES 
	%then %do;
		%let anno=;
		%if %length(&id) = 0 %then %do;
			%let id = _case_;
			%end;
		%label(data=&out, x=(expected-.2), y=dsq, text=&id, pos=4,
			subset=(_weight_=0));
			%let anno = anno=_label_;
			
		title2;
		proc gplot data=&out;
			plot dsq      * expected = 1
					expected * expected = 2
					/ overlay &anno hm=1 vm=1 vaxis=axis1
					  name="&name"
					  des="Outlier plot of &data";
			axis1 label=(a=90);
			symbol1 v=&symbol i=none c=black;
			symbol2 v=none i=join c=red;
			run;
	%end;
	
%done:
options nonotes;
proc datasets nofs nolist;
   delete outlier out1 trimmed;
	run; quit;
options notes;
%mend outlier;
