 /*-------------------------------------------------------------------*
  *    Name: TWOWAY.SAS                                               *
  *   Title: Analysis of two way tables                               *
  *     Doc: http://www.math.yorku.ca/SCS/sssg/twoway.html            *
  *                (Version 6+ only)                                  *
  * Macro for analysis of two way tables, including Tukey's           *
  * 1df test for non-additivity, and graphical display of             *
  * additive fit, together with a diagnostic plot for                 *
  * removable non-additivity.                                         *
  *-------------------------------------------------------------------*
  *  Author:  Michael Friendly            <friendly@YorkU.ca>         *
  * Created:  11 Dec 1989 10:51:22                                    *
  * Revised:  22 Oct 1997 11:39:26                                    *
  * Version:  1.6                                                     *
  *   - Changed to allow glm-style input                              *
  *   - Fixed problem with scale labels on vertical axis              *
  *   - Added afactor, rfactor options, arrow heads for resids        *
  *   - Added power transformation parameter                          *
  *                                                                   *
  *      From ``SAS System for Statistical Graphics, First Edition''  *
  *      Copyright(c) 1991 by SAS Institute Inc., Cary, NC, USA       *
  *-------------------------------------------------------------------*/

 /*
 Usage:
 
 The original version of this macro required that the columns of the
 data table be stored as a set of variables in the input dataset.  In
 this arrangment, use the VAR= argument to specify this list of variables
 and the ID= variable to specify an additional variable whose values
 are labels for the rows.  In this case, the RESPONSE= option is simply
 a character label for the response variable in the two-way plot.
 
 Assume a dataset of reaction times to 4 topics in 3 experimental tasks,
 in a SAS dataset like this:
     TASK   TOPIC1   TOPIC2   TOPIC3   TOPIC4
	  Easy     2.43     3.12     3.68     4.04
	  Medium   3.41     3.91     4.07     5.10
	  Hard     4.21     4.65     5.87     5.69
	     
 For this arrangment, the macro would be invoked as follows:
   %twoway(var=topic1 topic2 topic3 topic4, id=task, 
           response=Reaction Time);

 The present verion also allows the dataset to contain all response values
 in a single variable, with two additional variables to specify the row
 and column class variables, as is done with PROC GLM in the univariate
 (non-repeated measures) format. In this case, DO NOT specify an ID=
 variable, use the VAR= argument to specify the two row and column class 
 variables, and specify the name of response variable as RESPONSE=.
 
 The same data in this format would have 12 observations, and look like:
 		TASK  TOPIC    RT
		Easy    1     2.43
		Easy    2     3.12
		Easy    3     3.68
		...
		Hard    4     5.69
		
 For this arrangment, the macro would be invoked as follows:
   %twoway(var=topic task, response=RT);

 In this arrangement, the order of the VAR= variables does not matter.
 The columns of the two-way table are determined by the variable which
 varies most rapidly in the input dataset (topic, in the example).
 */
  
%macro TWOWAY(
        data=_LAST_,        /* Data set to be analyzed            */
        var=,               /* list of variables: cols of table   */
        id=,                /* row identifier: char variable      */
        response=Response,  /* Label for response on 2way plot    */
		  power=1,            /* Power transform of response        */
        plot=FIT DIAGNOSE,  /* What plots to do?                  */
		  htext=1.5,          /* text height for FIT plot           */
		  afactor=1,          /* angle factor for FIT plot labels   */
		  rfactor=1,          /* plot resids > rfactor*sqrt(MSPE)   */
		  rcolor=BLUE RED,    /* colors for + and - residuals       */
		  rline=3,            /* line style for residuals           */
		  arrow = .03,
        name=TWOWAY,        /* Name for graphic catalog plots     */
        gout=GSEG);         /* Name for graphic catalog           */
 
%if &var = %str() %then %do;
    %put ERROR: You must supply a VAR= variable list for columns;
    %goto DONE;
%end;

%if &id = %str() %then %do;
    %if %bquote(%scan(&var,2,%str( ))) = %str() %then %do;
    %put ERROR: When no ID= variable is specified, you must supply
	      two VAR= variable names, and the name of the RESPONSE=
			variable.;
    %goto DONE;
	 %end;
%end;

%if &sysver>6.10 %then %do;
	%*-- Find horizontal/vertical size: x.xx IN;
   %let hso=%sysfunc(getoption(hsize));
   %let vso=%sysfunc(getoption(vsize));
	%let hu = %scan(&hso,2);
	%let hs = %scan(&hso,1);
	%let vu = %scan(&vso,2);
	%let vs = %scan(&vso,1);
	%*-- Make the plot square;
	%if &hs > &vs %then  %let hs = &vs;
		%else %let vs = &hs;
	goptions hsize=&hs &hu vsize=&vs &vu;
   %end;

%let plot = %upcase(&plot);
%if &data=_LAST_ %then %let data=&syslast;

proc iml;
   reset;

start twoway;
 /*-------------------------------------------------------------*
  | Calculate points for lines in two-way display of fitted     |
  | value. Each point is (COLFIT+ROWEFF, COLFIT-ROWEFF).        |
  *-------------------------------------------------------------*/
    do i=1 to r;
       clo  = coleff[><]+all;
       from = from // (clo-roweff[i] || clo+roweff[i]);
       chi  = coleff[<>]+all;
       to   = to   // (chi-roweff[i] || chi+roweff[i]);
       labl = labl || rl[i];
       end;
    do j=1 to c;
       rlo  = roweff[><];
       to   = to   // (coleff[j]+all-rlo || coleff[j]+all+rlo);
       rhi  = roweff[<>];
       from = from // (coleff[j]+all-rhi || coleff[j]+all+rhi);
       labl = labl || cl[j];
       end;
 
    /*-----------------------------------*
     | Draw vectors for large residuals  |
     *-----------------------------------*/
    ahead = &arrow # ( max(fit)-min(fit) );
	 ahead = ahead # {-.25 -1, .25 -1};
    do i=1 to r;
    do j=1 to c;
       if abs(e[i, j]) > &rfactor # sqrt(mspe) then do;
          bot = ((cf[i,j]-re[i,j])||(cf[i,j]+re[i,j]));
			 top = bot + (0 || e[i,j]);
			*-- not drawn, but keep for scaling;
			 from = from // bot;
			 to   = to   // top;
			 ah   = sign(e[i,j])#ahead;
			 if e[i,j]>0 then do;
					frpos = frpos // bot // top // top;
					topos = topos // top // (top+ah[1,]) // (top+ah[2,]);
			 end;
			 else do;
					frneg = frneg // bot // top // top;
					toneg = toneg // top // (top+ah[1,]) // (top+ah[2,]);
			 end;
*			 from = from // top // top;
*			 to   = to // (top+ah[1,]) // (top+ah[2,]);
          end;
       end; end;
 
    /*----------------------------------*
     | Start IML graphics               |
     *---------------------------------**/
    %if &sysver < 6 %then %do;
       %let lib=%scan(&gout,1,.);
       %let cat=%scan(&gout,2,.);
       %if &cat=%str() %then %do;
           %let cat=&lib;
           %let lib=work;
           %end;
       call gstart gout={&lib &cat}
             name="&name" descript="Two-way plot for dataset &data";
    %end;
    %else %do;     /* Version 6 */
       call gstart("&gout");
       call gopen("&name",1,"Two-way plot of &response in &data");
    %end;
 
    /**--------------------------------**
     | Find scales for the two-way plot |
     **--------------------------------**/
/*
    call gport({10 10, 90 90});
    call gyaxis( {10 10}, 80, 5, 0,,'5.0') ;
    call gscale( scale2,from[,2]//to[,2], 5);
    call gscript(3, 40,"&Response",90,,3);
 
    call gscale( scale1, from[,1]//to[,1], 5);
    window = scale1[1:2] || scale2[1:2];
    call gwindow(window);
*/
	xx = from[,1]//to[,1];
	yy = from[,2]//to[,2];
	call gxyscale( xx, yy, {0 1}, { ' ' "&Response"}, ' ');
    /*----------------------------------*
     | Draw lines for fit and residuals |
     *----------------------------------*/
    l = nrow(from);
    call gdrawl( from[1:r+c,],   to[1:r+c,],1,"black");
	 c1 = scan("&rcolor &rcolor",1);
	 c2 = scan("&rcolor &rcolor",2);
    call gdrawl( frpos, topos,&rline, c1);
    call gdrawl( frneg, toneg,&rline, c2);
 
    /*----------------------------------------*
     | Plot row and column labels at margins; |
     *----------------------------------------*/
    xoffset=.04 * (to[<>,1]-to[><,1]);
    yoffset=0;
	 
    do i=1 to r+c;
	    angle = 45 * &afactor;
       if i>r then do;
          yoffset=-.04 * (to[<>,2]-to[><,2]);
			 angle = -45 * &afactor;
          end;
       call gscript(xoffset+to[i,1],yoffset+to[i,2],labl[i],angle,,&htext);
    end;
    call gshow;
    call gstop;
	finish;

 /*-------------------------------------------------------------*/
 /*-- Sets the range of data allowing for p1% of space below, --*/
 /*-- and p2% of space above. ----------------------------------*/
 /*-------------------------------------------------------------*/
start rngset(w, s, p1, p2);
   call gscale(s,s[1:2], s[3]);
   w[1] = s[2] - s[1];         /*-- rescaled range ------------*/
   s[3] = w[1]/s[3];           /*-- # of interval -------------*/
   w[2] = w[1]/(1-p1-p2);      /*-- expand range --------------*/
   w[1] = s[1] - p1*w[2];      /*-- move lower bound ----------*/
   w[2] = s[2] + p2*w[2];      /*-- move upper edge -----------*/
finish;
 
start gxyscale(x,y, axes, label, title)
		global(xticks, yticks, caxis, ctext, ftext, htext, xscale, yscale);

	if type(xticks)='U' then xticks=5;		
	if type(yticks)='U' then yticks=5;		
	if type(caxis)='U' then caxis='BLACK';		
	if type(ctext)='U' then ctext='BLACK';		
	if type(ftext)='U' then ftext='DUPLEX';		
	if type(htext)='U' then htext=&htext;
	
   /*-- set window --------------------------------------------*/
   w = {0 0, 100 100};  origin = {0,0};
   xscale = x[><] // x[<>] // xticks;
   run rngset( origin, xscale, .1, .05);
   w[,1] = origin;
   yscale = y[><] // y[<>] // yticks;
   run rngset( origin, yscale, .1, .05);
   w[,2] = origin;   origin = xscale[1] || yscale[1];
   call gwindow(w);
	* print 'Xscale, Yscale' xscale yscale;
	* print 'window' w;

	if all(mod(xscale,1)=0) then xfmt='6.0';
								   else xfmt='7.1';
	if all(mod(yscale,1)=0) then yfmt='6.0';
								   else yfmt='7.1';
 
   /*-- draw axes and axis labels -----------------------------*/

	if axes[1] ^=0 then do;
		call gxaxis(origin,xscale[2]-xscale[1],xscale[3],,,xfmt,,,caxis,'N');
		call gstrlen(len, label[1]);
		cx = (xscale[2]+xscale[1]-len)/2;
		call gscript(cx, 8, label[1],,,htext,,,w[,1] || {0,100});
		end;
	
	if axes[2] ^=0 then do;
		call gyaxis(origin,yscale[2]-yscale[1],yscale[3],,,yfmt,,,caxis,'N');
		call gstrlen(len, label[2], htext+.75);
		cy = (yscale[2]+yscale[1]-len)/2;
		call gscript(4, cy, label[2],90,,htext+.75,,,{0,100} || w[,2]);
		end;

   /*-- do title ----------------------------------------------*/
   if (substr(title,1,1) ^= ' ') then do;
      call gstrlen(len, title, htext, ftext);
		cx = (xscale[2]+xscale[1]-len)/2;
      call gscript(cx,(w[2,2]+yscale[2])/2,
                  title,0,0,htext,ftext,ctext);
      end;
finish;
 
/* --------------------------------------------------------------------
 Routine to read frequency and index/label variables from a SAS dataset
 and construct the appropriate levels, and lnames variables

 Input:  dataset - name of SAS dataset (e.g., 'mydata' or 'lib.nydata')
         variable - name of variable containing frequencies
		   vnames - character vector of names of index variables
 Output: dim (numeric levels vector)
         lnames (K x max(dim)) 
  --------------------------------------------------------------------*/
 
start readtab(dataset, variable, vnames, table, dim, lnames);
	if type(vnames)^='C'
		then do;
			print 'VNAMES argument must be a character vector';
			show vnames;
			return;
		end;		
   if nrow(vnames)=1 then vnames=vnames`;

	call execute('use ', dataset, ';');
	read all var variable into table;
	run readlab(dim, lnames, vnames);
	call execute('close ', dataset, ';');
	reset noname;
	print 'Variable' variable 'read from dataset' dataset,
		'with factors ordered' vnames lnames;
	reset name;
	finish;

/* Read variable index labels from an open dataset, construct a dim 
   vector and lnames matrix so that variables are ordered correctly
   for mosaics and ipf (first varying most rapidly).

	The data set is assumed to be sorted by all index variables.  If
	the observations were sorted by A B C, the output will place
	C first, then B, then A.
   Input:    vnames (character K-vector)
 */

start readlab(  dim, lnames, vnames);
	free span lnames dim;
	nv = nrow(vnames);

	spc = '        ';
	do i=1 to nv;
		vi = vnames[i,];
		read all var vi into cli;
		if type(cli) = 'N'
			then do;
				tmp = trim(left(char(cli,8)));
				tmp = substr(tmp,1,max(length(tmp)));
				cli = tmp;
				end; 
		cli = trim(cli); 
		span = span || loc(0=(cli[1,] = cli))[1];
		d=design( cli );
		dim = dim || ncol(d);
		free row1;
		*-- find position of each first distinct value;
		do j=1 to ncol(d);
			row1 = row1 || loc(d[,j]=1)[1];
			end;
		*-- sort elements in row1 so that var labels are in data order;
		order = rank(row1);
		tmp = row1;
		row1[,order]=tmp;	

		li = t(cli[row1]);
		if i=1 then lnames = li;
			else do;
				if ncol(lnames) < ncol(row1) 
					then lnames=lnames || repeat(spc, i-1, ncol(row1)-ncol(lnames));
				if ncol(lnames) > ncol(row1)
					then li = li || repeat(spc, 1, ncol(lnames)-ncol(li));
				lnames = lnames // li;
			end;
		end;

	*-- sort index variables by span so that last varies most slowly;
	order = rank(span);
	tmp = span;
	span[,order] = tmp;
	tmp = dim;
	dim[,order] = tmp;
	tmp = lnames;
	lnames[order,] = tmp;
	tmp = vnames;
	vnames[order,] = tmp;
	*	print dim lnames vnames;
	finish;
 
start power(x, pow);
		if pow=1 then return(x);
		if any(x <= 0) then x = x + ceil(min(x)+.5);
		if abs(pow)<.001 then xt =  log(x);
			else xt = ((x##pow)-1) / pow;
	return (xt);
	finish;

	/*--- Main routine */
	%if &id = %str() %then %do;
		run readtab("&data", "&response", { &var}, y, dim, lnames);
		y = shape(y,dim[1], dim[2]);
		clabel = lnames[1,];
		rlabel = lnames[2,];
		%end;
   %else %do; 
	  use &data;
     read all into  y[colname=clabel rowname=&id] var { &var };
	  rlabel = t(&id);
	  %end;

	%if &power ^= 1 %then %do;
		y = power(y, &power);
		%end;

     r = nrow( y);
     c = ncol( y);
     rowmean = y[ , :];
     colmean = y[ : ,];
     all = y[ :  ];                        * grand mean ;
 
     roweff = rowmean - all;                 * row effects;
     coleff = colmean - all;                 * col effects;
     data = ( y       || round(rowmean,.001) || round(roweff,.001) ) //
            ( round(colmean,.001) || round(all,.001) || 0 ) //
            ( round(coleff,.001)  || 0       || 0 );
 
     rl    = rlabel || {'COLMEAN' 'COLEFF'};
     cl    = clabel || {'ROWMEAN' 'ROWEFF'};
     print , data [ rowname = rl colname = cl ];
 
     jc = j( r , 1);
     jr = j( 1 , c);
     e = y - (rowmean * jr) - (jc * colmean) + all;
	  fit = y - e;
     print 'Interaction Residuals ',
           e [ rowname = rl colname = cl format=8.3 ];
 
     sse = e[ ## ];
     dfe = ( r - 1 ) # ( c - 1 );
 
     ssrow = roweff[## ,];  ssa = c * ssrow;
     sscol = coleff[ ,##];  ssb = r * sscol;
     product = ( roweff * coleff ) # y;
     d = product[ + ]  / ( ssrow # sscol );
 
     ssnon = ( ( product[ + ] ) ## 2 ) / ( ssrow # sscol );
     sspe = sse - ssnon;
     ss   = ssa   // ssb // sse // ssnon // sspe ;
     df   = (r-1) //(c-1)// dfe //  1    // dfe-1;
 
     ms = ss / df ; mspe=sspe/(dfe-1);
     f  = ms / (ms[{3 3 3 5 5},]);
 
     source=  { "Rows","Cols","Error","Non-Add","Pure Err"};
     srt = "SOURCE   ";
     sst = "   SS    ";
     dft = "   DF    ";
     mst = "   MS    ";
     ft  = "   F     ";
     reset noname;
     print "Analysis of Variance Summary Table for &Response"
	  		%if &power ^=1 %then %str(" ## &power");
	  			,
           'with Tukey 1 df test for Non - Additivity ',,
           source[ colname=srt ]
           ss[ colname=sst format=9.3]  df[ colname=dft format=best8. ]
           ms[ colname=mst format=9.3]  f [ colname=ft];
 
     re = ( roweff * jr );
     cf = ( jc * coleff )  + all;
 
     compare = ( roweff * coleff ) / all;
     compare = shape( e ,0 , 1) ||
               shape( compare ,0 , 1) ||
               shape( re ,0 ,1) ||
               shape( cf ,0, 1);
     vl = { 'RESIDUAL' 'COMPARE' 'ROWEFF' 'COLFIT'};
     create compare from compare[ colname=vl ];
     append from compare;
 
     /* Calculate slope of Residuals on Comparison values */
     /* for possible power transformation                 */
     xy = compare[,{2 1}];
     slope = sum(xy[,1] # xy[,2]) / xy[##,1];
     slope = d || slope || (1-slope);
     D1={"D" "Slope" "Power"};
     print 'D = Coefficient of alpha ( i ) * beta ( j ) ' ,
           'Slope of regression of Residuals on Comparison values',
           '1 - slope = power for transformation',,
           slope[ colname=D1];
    ;
    %if %index(&plot,FIT) > 0 %then %do;
    run twoway;
    %end;
 
quit;
 
%if &sysver>6.10 %then %do;
	%*-- Restore original plot size;
	goptions hsize=&hso vsize=&vso;
   %end;

data compare;
   set compare;
   fit = colfit + roweff;
   data= fit + residual;
   diff= colfit - roweff;
 
   /* Print values for fit and diagnostic plots */
proc print data=compare;
   var data roweff colfit fit diff residual compare;
%if %index(&plot,PRINT) > 0 %then %do;
proc plot;
     plot data* diff = '+'
          fit * diff = '*'    / overlay;
proc plot;
     plot residual * compare / vpos=45;
%end;
 
%if %index(&plot,DIAGNOSE) > 0 %then %do;
proc gplot data=compare gout=&gout;
     plot residual * compare
        / vaxis=axis1    /*haxis=axis2*/
          vminor=1 hminor=1 name="&name"
          des="Two-way diagnostic plot for &data";
     symbol1 v=- h=1.4 c=black i=rl;
     axis1 label=(a=90 r=0);
*     axis2 label=(h=1.5);
     label residual = 'Interaction Residual'
           compare  = 'Comparison Value';
run;
%end;
 
%DONE:
%mend TWOWAY;
